package me.hammerle.code;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Stack;
import java.util.TreeSet;
import me.hammerle.exceptions.PrescriptException;

public class Code implements Comparable<Code>
{
    private final int line;
    private int subline;
    
    private final String function;
    private final int level;
    private final int pars;
    
    private final Object value;
    private int jump;
    
    public Code(String function, int level, int pars, int line, int subline, Object value, int jump)
    {
        if(function != null)
        {
            this.function = function.toLowerCase();
        }
        else
        {
            this.function = null;
        }
        this.level = level;
        this.pars = pars;
        this.line = line;
        this.subline = subline;
        this.value = value;
        this.jump = jump;
    }
    
    public Code(String function, int level, int pars, int line, int subline)
    {
        this(function.trim(), level, pars, line, subline, null, 0);
    }
    
    public Code(int level, int pars, int line, int subline, Object value)
    {
        this(null, level, pars, line, subline, value, 0);
    }
    
    public void executeFunction(Script sc, Stack<Object> stack)
    {
        //System.out.println("AUSFÜHRUNG: " + this.toString());
        if(value != null)
        {
            if(value.getClass() == Variable.class)
            {
                stack.push(sc.getVar(((Variable) value).getName()));
                return;
            }
            stack.push(value);
            return;
        }
        Object[] input;
        if(pars > 0)
        {
            input = new Object[pars];
            for(int i = 0; i < pars; i++)
            {
                input[i] = stack.pop();
            }
        }
        else
        {
            input = new Object[0];
        }
        Object output = CodeParser.parseFunction(sc, function, input);
        if(output != Void.TYPE)
        {
            stack.push(output);
        }
    }
    
    public String getFunction()
    {
        return function;
    }

    @Override
    public String toString() 
    {
        StringBuilder sb = new StringBuilder(">>>");
        for(int i = 1; i < level; i++)
        {
            sb.append(">>>");
        }
        sb.append(" (");
        sb.append(line);
        sb.append(",");
        sb.append(subline);
        sb.append(") ");
        
        if(value != null)
        {
            sb.append("push ");
            sb.append(value.getClass());
            sb.append(" value = '");
            sb.append(value);
            sb.append("'");
            return sb.toString();
        }
        
        sb.append("function ");
        sb.append(function);
        sb.append(" (");
        for(int i = 0; i < pars; i++)
        {
            sb.append("X,");
        }
        if(pars > 0)
        {
            sb.deleteCharAt(sb.length() - 1);
        }
        sb.append(")");
        if(jump != 0)
        {
            sb.append(" (");
            sb.append(jump);
            sb.append(")");
        }
        return sb.toString();
    }
    
    @Override
    public int compareTo(Code o) 
    {
        int i = Integer.compare(line, o.line);
        if(i == 0)
        {
            return Integer.compare(o.subline, subline);
        }
        return i;
    }
    
    public int getJumpLine()
    {
        return jump;
    }
    
    // -----------------------------------------------------------------------------------
    // Code-Builder
    // -----------------------------------------------------------------------------------
    
    private static int sublines = 0;
    
    private static int findEndOfLine(String code, int pos)
    {
        int start = pos;
        int length = code.length();
        char c;
        boolean text = false;
        while(pos < length)
        {
            c = code.charAt(pos);
            if(text && c != '"')
            {
                pos++;
                continue;
            }
            switch(c)
            {
                case ';':
                case '{':
                case '}':
                    return pos;
                case '"':
                    text = !text;
                    break;
            }
            pos++;
        }
        return start;
    }
    
    private static int findStartOfSyntax(String code, int pos)
    {
        int bracketCounter = 0;
        char c;
        boolean text = false;
        while(code.charAt(pos) == ' ')
        {
            pos--;
        }
        while(pos >= 0)
        {
            c = code.charAt(pos);
            if(text && c != '"')
            {
                pos--;
                continue;
            }
            switch(c)
            {
                case ';':
                case '{':
                case '}':
                    if(bracketCounter != 0)
                    {
                        throw new PrescriptException("unbalanced ()", code.substring(pos - 1, Math.min(code.length(), pos + 10))); 
                    }
                    return pos + 1;
                case ' ':
                case '+':
                case '-':
                case '*':
                case '/':
                case '^':
                case '=':
                case '>':
                case '<':
                case '!':
                case '%':
                case ',':
                case '&':
                case '|':
                    if(bracketCounter != 0)
                    {
                        break;
                    }
                    return pos + 1;
                case '"':
                    text = !text;
                    break;
                case ')':
                    bracketCounter++;
                    break;
                case '(':
                    if(bracketCounter == 0)
                    {
                        return pos + 1;
                    }
                    bracketCounter--;
                    break;
            }
            pos--;
        }
        return 0;
    }
    
    private static int findEndOfSyntax(String code, int pos)
    {
        int bracketCounter = 0;
        char c;
        boolean text = false;
        while(code.charAt(pos) == ' ')
        {
            pos++;
        }
        while(pos < code.length())
        {
            c = code.charAt(pos);
            if(text && c != '"')
            {
                pos++;
                continue;
            }
            switch(c)
            {
                case ';':
                case '{':
                case '}':
                    if(bracketCounter != 0)
                    {
                        throw new PrescriptException("unbalanced ()", code.substring(pos - 1, Math.min(code.length(), pos + 10))); 
                    }
                    return pos;
                case ' ':
                case '+':
                case '-':
                case '*':
                case '/':
                case '^':
                case '=':
                case '>':
                case '<':
                case '!':
                case '%':
                case ',':
                case '&':
                case '|':
                    if(bracketCounter != 0)
                    {
                        break;
                    }
                    return pos;
                case '"':
                    text = !text;
                    break;
                case '(':
                    bracketCounter++;
                    break;
                case ')':
                    if(bracketCounter == 0)
                    {
                        return pos;
                    }
                    bracketCounter--;
                    break;
            }
            pos++;
        }
        return code.length();
    }
    
    private static int findSyntax(String find, String code, int pos)
    {
        int length = code.length();
        char c;
        boolean text = false;
        String s;
        while(pos < length)
        {
            c = code.charAt(pos);
            if(text && c != '"')
            {
                pos++;
                continue;
            }
            if(c == '"')
            {
                text = !text;
                pos++;
                continue;
            }
            s = code.substring(pos);
            if(s.startsWith(find) && !s.startsWith(find + "=") && code.charAt(pos - 1) != '=')
            {
                return pos;
            }
            pos++;
        }
        return -1;
    }
    
    private static String changeBiSyntax(String syntax, String f, String code, boolean b)
    {
        int pos = -1;
        int end;
        int start;
        int slength = syntax.length();
        StringBuilder sb = new StringBuilder(code);
        StringBuilder newSyntax;
        while(pos < sb.length())
        {
            pos = findSyntax(syntax, sb.toString(), pos + 1);
            if(pos == -1)
            {
                break;
            }
            start = findStartOfSyntax(sb.toString(), pos - 1);
            end = findEndOfSyntax(sb.toString(), pos + slength);
            newSyntax = new StringBuilder(f);
            newSyntax.append("(");
            if(b)
            {
                newSyntax.append("\"");
            }
            newSyntax.append(sb.substring(start, pos).trim());
            if(b)
            {
                newSyntax.append("\"");
            }
            if(end > pos)
            {
                newSyntax.append(",");
                newSyntax.append(sb.substring(pos + slength, end).trim());
            }
            newSyntax.append(")");
            pos -= end - start - newSyntax.length();
            //System.out.println(sb.substring(start, end) + "   ===>    " + newSyntax);
            sb.replace(start, end, newSyntax.toString());
        }
        return sb.toString();
    }
    
    private static String changeUnSyntax(String syntax, String f, String code)
    {
        int pos = -1;
        int end;
        int start;
        int slength = syntax.length();
        String first;
        StringBuilder sb = new StringBuilder(code);
        StringBuilder newSyntax;
        while(pos < sb.length())
        {
            pos = findSyntax(syntax, sb.toString(), pos + 1);
            if(pos == -1)
            {
                break;
            }
            start = findStartOfSyntax(sb.toString(), pos - 1);
            end = findEndOfSyntax(sb.toString(), pos + slength);
            newSyntax = new StringBuilder("setvar(\"");
            first = sb.substring(start, pos).trim();
            newSyntax.append(first);
            newSyntax.append("\",");
            newSyntax.append(f);
            newSyntax.append("(");
            newSyntax.append(first);
            newSyntax.append(",");
            newSyntax.append(sb.substring(pos + slength, end).trim());
            newSyntax.append("))");
            pos -= end - start - newSyntax.length();
            sb.replace(start, end, newSyntax.toString());
        }
        return sb.toString();
    }
    
    private static String replaceChar(char find, char replacement, String code)
    {
        char[] chars = code.toCharArray();
        int length = code.length();
        char c;
        boolean text = false;
        int pos = 0;
        while(pos < length)
        {
            c = code.charAt(pos);
            if(text && c != '"')
            {
                pos++;
                continue;
            }
            if(c == '"')
            {
                text = !text;
                pos++;
                continue;
            }
            if(c == find)
            {
                chars[pos] = replacement;
            }
            pos++;
        }
        return new String(chars);
    }
    
    public static Code[] generate(String code, HashMap<String, Integer> gotos)
    {
        try
        {
            // Anfängliches Austauschen mit Functionssyntax
            code = changeBiSyntax("^", "math.pow", code, false);
            code = changeBiSyntax("/", "div", code, false);
            code = changeBiSyntax("%", "math.mod", code, false);
            code = changeBiSyntax("*", "mul", code, false);
            code = changeBiSyntax("-", "sub", code, false);
            code = changeBiSyntax("+", "add", code, false);

            code = changeBiSyntax("==", "equal", code, false);
            code = changeBiSyntax("!=", "notequal", code, false);
            code = changeBiSyntax("<", "less", code, false);
            code = changeBiSyntax(">", "greater", code, false);
            code = changeBiSyntax(">=", "greaterequal", code, false);
            code = changeBiSyntax("<=", "lessequal", code, false);
            
            code = changeBiSyntax("||", "or", code, false);
            code = changeBiSyntax("&&", "and", code, false);

            code = changeUnSyntax("+=", "add", code);
            code = changeUnSyntax("-=", "sub", code);
            code = changeUnSyntax("/=", "div", code);
            code = changeUnSyntax("*=", "mul", code);

            code = changeBiSyntax("=", "setvar", code, true);

            //code = replaceChar('[', '(', code);
            //code = replaceChar(']', ')', code);

            //System.out.println(code);
            //System.exit(0);
            // Ende des Austauschen
            // Zerlegen des Codes in Zeilen
            ArrayList<Code> list = new ArrayList<>();
            sublines = 0;
            String actual;
            int length = code.length();
            int level = 1;
            int old;
            int pos = 0;
            int line = 0;
            while(pos < length)
            {
                old = pos;
                pos = findEndOfLine(code, pos);
                actual = code.substring(old, pos).trim();
                //System.out.println("WUSI: " + actual);
                if(actual.startsWith("@"))
                {
                    gotos.put(actual.substring(1), line);
                    pos++;
                    continue;
                }
                switch(code.charAt(pos))
                {
                    case '{':
                        line++;
                        splitFunctions(list, actual, level, line);
                        level++;
                        break;
                    case '}': 
                        level--;
                        break;
                    case ';': 
                        line++;
                        splitFunctions(list, actual, level, line);
                        break;
                }
                pos++;
            }
            // Ende der Zeilenzerlegung

            // Anlegen eines Trees zum sortieren und einfügen von Code
            TreeSet<Code> tree = new TreeSet(list);
            Code[] helper = tree.toArray(new Code[tree.size()]);

            // Einfügen von "gotoline" bei while-Schleifen
            String function;
            int baseLevel;
            for(int i = 0; i < helper.length; i++)
            {
                function = helper[i].function;
                if(function == null || !function.equals("while"))
                {
                    continue;
                }
                baseLevel = helper[i].level;
                for(int j = i + 1; j < helper.length; j++)
                {
                    if(helper[j].level <= baseLevel)
                    {
                        helper[i].jump = j - i + 1;
                        Code c = new Code("gotoline", helper[i].level, 0, helper[j].line, helper[j].subline + 1);
                        while(tree.contains(c)) // Damit keine Whiles, die gleich enden, sich überschreiben
                        {
                            c.subline++;
                        }
                        tree.add(c);
                        break;
                    }
                }
            }
            // Ende des While-Handlings
            // Erstellen der Sprungmarken der Schlüsselwörter
            helper = tree.toArray(new Code[tree.size()]);

            boolean whileCheck = false;
            int lastLineChange = 0;
            line = 0;
            for(int i = 0; i < helper.length; i++)
            {
                function = helper[i].function;
                if(helper[i].line != line)
                {
                    line = helper[i].line;
                    lastLineChange = i;
                }
                if(function == null)
                {
                    continue;
                }
                switch(function)
                {
                    case "if":
                    case "else":
                    case "try":
                    case "catch":
                        break;
                    case "while":
                        whileCheck = true;
                        break;
                    default:
                        continue;
                }
                baseLevel = helper[i].level;
                for(int j = i + 1; j < helper.length; j++)
                {
                    if(helper[j].level <= baseLevel)
                    {
                        helper[i].jump = j - i;
                        if(whileCheck)
                        {
                            whileCheck = false;
                            helper[j].jump = lastLineChange - j;
                        }
                        break;
                    }
                }
                if(helper[i].jump == 0)
                {
                   helper[i].jump = list.size() - i; 
                }
            }
            // Ende der Srungmarken
            // Korrektur der Gotos;
            Code[] c = tree.toArray(new Code[list.size()]);
            gotos.entrySet().forEach((i) -> 
            {
                int value = i.getValue() + 1;
                for(int j = value; j < c.length; j++)
                {
                    if(c[j].line == value)
                    {
                        i.setValue(j);
                        return;
                    }
                }
            });
            // Ende
            //java.util.Arrays.stream(c).forEach(co -> System.out.println(co.toString()));
            //gotos.forEach((k, v) -> System.out.println(k + "   " + v));
            //System.exit(0);
            return c;
        }
        catch(PrescriptException ex)
        {
            ScriptUtils.printError("PrescriptException");
            ScriptUtils.printError(ex.getBadString() + " - " + ex.getException());
            return null;
        }
    } 
    
    private static int findOpenBracket(String code, int pos)
    {
        int length = code.length();
        char c;
        boolean text = false;
        while(pos < length)
        {
            c = code.charAt(pos);
            if(text && c != '"')
            {
                pos++;
                continue;
            }
            switch(c)
            {
                case '(':
                    return pos;
                case '"':
                    text = !text;
                    break;
            }
            pos++;
        }
        return -1;
    }
    
    private static int findClosingBracket(String code, int pos)
    {
        if(pos == -1)
        {
           pos++;
        }
        int length = code.length();
        char c;
        boolean text = false;
        int brackets = 0;
        while(pos < length)
        {
            c = code.charAt(pos);
            if(text && c != '"')
            {
                pos++;
                continue;
            }
            switch(c)
            {
                case '(':
                    brackets++;
                    break;
                case ')':
                    brackets--;
                    if(brackets == 0)
                    {
                        return pos;
                    }
                    break;
                case '"':
                    text = !text;
                    break;
            }
            pos++;
        }
        return -1;
    }
    
    private static ArrayList<String> splitComma(String code)
    {
        ArrayList<String> list = new ArrayList<>();
        int length = code.length();
        int pos = 0;
        int old = 0;
        int brackets = 0;
        char c;
        boolean text = false;
        while(pos < length)
        {
            c = code.charAt(pos);
            if(text && c != '"')
            {
                pos++;
                continue;
            }
            
            switch(c)
            {
                case '"':
                    text = !text;
                    break;
                case '(':
                    brackets++;
                    break;
                case ')':
                    brackets--;
                    break;
                case ',':
                    if(brackets == 0)
                    {
                        list.add(code.substring(old, pos));
                        old = pos + 1;
                    }
                    break;
            }
            pos++;
        }
        if(old < pos)
        {
            list.add(code.substring(old, pos));
        }
        return list;
    }
    
    private static void splitFunctions(ArrayList<Code> list, String f, int level, int line)
    {
        sublines++;
        int start = findOpenBracket(f, 0);
        int end = findClosingBracket(f, start);
        if((start != -1 || end != -1) && (((start == -1 || end == -1) && start != end) || end != f.length() - 1))
        {
            throw new PrescriptException("unbalanced ()", f);
        }
        if(start == -1)
        {
            list.add(new Code(level, 0, line, sublines, convertInput(f)));
            return;
        }
        else if(start >= end - 1)
        {
            list.add(new Code(f.substring(0, start), level, 0, line, sublines));
            return;
        }
        ArrayList<String> splitted = splitComma(f.substring(start + 1, end));
        list.add(new Code(f.substring(0, start), level, splitted.size(), line, sublines));
        splitted.forEach(s -> splitFunctions(list, s, level, line));
    }
    
    public static Object convertInput(String s)
    {
        if(s == null)
        {
            return null;
        }
        s = s.trim();
        if(s.startsWith("\"") && s.endsWith("\""))
        {
            if(s.length() <= 1)
            {
                return "\"";
            }
            return s.substring(1, s.length() - 1);
        }
        else if(s.equals("true"))
        {
            return true;
        }
        else if(s.equals("false"))
        {
            return false;
        }
        else if(s.equals("null"))
        {
            return null;
        }
        try
        {
            return Double.parseDouble(s);
        }
        catch(NumberFormatException ex)
        {
            return new Variable(s);
        }
    }
}
